@@ -90,11 +90,11 @@ import qualified Control.Exception as E
 import System.IO ( stderr, hPutStrLn )
 import qualified Data.Digest.CRC32 as CRC32
 import qualified Data.Map as M
-#if MIN_VERSION_binary(0,6,0)
 import Control.Applicative
-#endif
-#ifndef _WINDOWS
-import System.Posix.Files ( setFileTimes, setFileMode, fileMode, getSymbolicLinkStatus, symbolicLinkMode, readSymbolicLink, isSymbolicLink, unionFileModes, createSymbolicLink )
+#ifdef _WINDOWS
+import Data.Char (isLetter)
+#else
+import System.Posix.Files ( setFileTimes, setFileMode, fileMode, getSymbolicLinkStatus, symbolicLinkMode, readSymbolicLink, isSymbolicLink, unionFileModes, createSymbolicLink, removeLink )
 import System.Posix.Types ( CMode(..) )
 import Data.List (partition)
 import Data.Maybe (fromJust)
@@ -111,8 +111,8 @@ import qualified Data.Text.Lazy.Encoding as TL
 
 -- from zlib
 import qualified Codec.Compression.Zlib.Raw as Zlib
+import System.IO.Error (isAlreadyExistsError)
 
-#if !MIN_VERSION_binary(0, 6, 0)
 manySig :: Word32 -> Get a -> Get [a]
 manySig sig p = do
     sig' <- lookAhead getWord32le
@@ -122,7 +122,6 @@ manySig sig p = do
             rs <- manySig sig p
             return $ r : rs
         else return []
-#endif
 
 
 ------------------------------------------------------------------------
@@ -335,13 +334,32 @@ readEntry opts path = do
 
   when (OptVerbose `elem` opts) $ do
     let compmethod = case eCompressionMethod entryE of
-                     Deflate       -> "deflated"
+                     Deflate       -> ("deflated" :: String)
                      NoCompression -> "stored"
     hPutStrLn stderr $
       printf "  adding: %s (%s %.f%%)" (eRelativePath entryE)
       compmethod (100 - (100 * compressionRatio entryE))
   return entryE
 
+-- check path, resolving .. and . components, raising
+-- UnsafePath exception if this takes you outside of the root.
+checkPath :: FilePath -> IO ()
+checkPath fp =
+  maybe (E.throwIO (UnsafePath fp)) (\_ -> return ())
+    (resolve . splitDirectories $ fp)
+  where
+    resolve =
+      fmap reverse . foldl go (return [])
+      where
+      go acc x = do
+        xs <- acc
+        case x of
+          "."  -> return xs
+          ".." -> case xs of
+                    []     -> fail "outside of root path"
+                    (_:ys) -> return ys
+          _    -> return (x:xs)
+
 -- | Writes contents of an 'Entry' to a file.  Throws a
 -- 'CRC32Mismatch' exception if the CRC32 checksum for the entry
 -- does not match the uncompressed data.
@@ -350,15 +368,11 @@ writeEntry opts entry = do
   when (isEncryptedEntry entry) $
     E.throwIO $ CannotWriteEncryptedEntry (eRelativePath entry)
   let relpath = eRelativePath entry
-  let isUnsafePath = ".." `elem` splitDirectories relpath
-  when isUnsafePath $
-    E.throwIO $ UnsafePath relpath
+  checkPath relpath
   path <- case [d | OptDestination d <- opts] of
-             (x:_) -> return (x </> relpath)
-             _ | isAbsolute relpath
-                   -> E.throwIO $ UnsafePath relpath
-               | otherwise
-                   -> return relpath
+             (x:_)                   -> return (x </> relpath)
+             [] | isAbsolute relpath -> E.throwIO $ UnsafePath relpath
+                | otherwise          -> return relpath
   -- create directories if needed
   let dir = takeDirectory path
   exists <- doesDirectoryExist dir
@@ -405,10 +419,18 @@ writeSymbolicLinkEntry opts entry =
              let symlinkPath = prefixPath </> eRelativePath entry
              when (OptVerbose `elem` opts) $ do
                hPutStrLn stderr $ "linking " ++ symlinkPath ++ " to " ++ targetPath
-             createSymbolicLink targetPath symlinkPath
+             forceSymLink targetPath symlinkPath
            else writeEntry opts entry
 
 
+-- | Writes a symbolic link, but removes any conflicting files and retries if necessary.
+forceSymLink :: FilePath -> FilePath -> IO ()
+forceSymLink target linkName =
+    createSymbolicLink target linkName `E.catch`
+      (\e -> if isAlreadyExistsError e
+             then removeLink linkName >> createSymbolicLink target linkName
+             else ioError e)
+
 -- | Get the target of a 'Entry' representing a symbolic link. This might fail
 -- if the 'Entry' does not represent a symbolic link
 symbolicLinkEntryTarget :: Entry -> Maybe FilePath
@@ -467,7 +489,13 @@ normalizePath :: FilePath -> String
 normalizePath path =
   let dir   = takeDirectory path
       fn    = takeFileName path
-      (_drive, dir') = splitDrive dir
+      dir' = case dir of
+#ifdef _WINDOWS
+               (c:':':d:xs) | isLetter c
+                            , d == '/' || d == '\\'
+                            -> xs  -- remove drive
+#endif
+               _ -> dir
       -- note: some versions of filepath return ["."] if no dir
       dirParts = filter (/=".") $ splitDirectories dir'
   in  intercalate "/" (dirParts ++ [fn])
@@ -661,15 +689,9 @@ setFileTimeStamp file epochtime = do
 
 getArchive :: Get Archive
 getArchive = do
-#if MIN_VERSION_binary(0,6,0)
-  locals <- many getLocalFile
-  files <- many (getFileHeader (M.fromList locals))
-  digSig <- Just `fmap` getDigitalSignature <|> return Nothing
-#else
   locals <- manySig 0x04034b50 getLocalFile
   files <- manySig 0x02014b50 (getFileHeader (M.fromList locals))
-  digSig <- lookAheadM getDigitalSignature
-#endif
+  digSig <- Just `fmap` getDigitalSignature <|> return Nothing
   endSig <- getWord32le
   unless (endSig == 0x06054b50)
     $ fail "Did not find end of central directory signature"
@@ -776,7 +798,6 @@ getLocalFile = do
   return (fromIntegral offset, compressedData)
 
 getWordsTilSig :: Word32 -> Get B.ByteString
-#if MIN_VERSION_binary(0, 6, 0)
 getWordsTilSig sig = (B.fromChunks . reverse) `fmap` go Nothing []
   where
     sig' = S.pack [fromIntegral $ sig .&. 0xFF,
@@ -824,17 +845,6 @@ getWordsTilSig sig = (B.fromChunks . reverse) `fmap` go Nothing []
           where
             len = S.length chunk
             prefixes' = Just $ (S.index chunk (len - 3), S.index chunk (len - 2), S.index chunk (len - 1))
-#else
-getWordsTilSig sig = B.pack `fmap` go []
-  where
-    go acc = do
-      sig' <- lookAhead getWord32le
-      if sig == sig'
-          then skip 4 >> return (reverse acc)
-          else do
-              w <- getWord8
-              go (w:acc)
-#endif
 
 putLocalFile :: Entry -> Put
 putLocalFile f = do
@@ -974,22 +984,11 @@ putFileHeader offset local = do
 -- >     size of data                    2 bytes
 -- >     signature data (variable size)
 
-#if MIN_VERSION_binary(0,6,0)
 getDigitalSignature :: Get B.ByteString
 getDigitalSignature = do
   getWord32le >>= ensure (== 0x05054b50)
   sigSize <- getWord16le
   getLazyByteString (toEnum $ fromEnum sigSize)
-#else
-getDigitalSignature :: Get (Maybe B.ByteString)
-getDigitalSignature = do
-  hdrSig <- getWord32le
-  if hdrSig /= 0x05054b50
-     then return Nothing
-     else do
-        sigSize <- getWord16le
-        getLazyByteString (toEnum $ fromEnum sigSize) >>= return . Just
-#endif
 
 putDigitalSignature :: Maybe B.ByteString -> Put
 putDigitalSignature Nothing = return ()
